/*

 x1541.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "x1541.h"
#include <errno.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#elif defined __FreeBSD__
#include <fcntl.h>
#elif defined __DJGPP__
#else
#include <sys/io.h>
#endif

#include <unistd.h>

#include "crc8.h"
#include "ideservd.h"
#include "log.h"

static unsigned short int IOPORT, STATPORT, CTRLPORT, PORTIN;
static int inv;
static const char *i_dev;
static int i_port;
static int inited;
static unsigned char DATALO, DATAHI, CLKLO, CLKHI, ATNHI, DATA_, CLK_, ATN_, DATA;
#define CLK 2

#ifdef WIN32
/* prototype (function typedef) for DLL function Inp32: */

typedef short _stdcall (*inpfuncPtr)(short portaddr);
typedef void _stdcall (*oupfuncPtr)(short portaddr, short datum);

static HINSTANCE hLib;
static inpfuncPtr inp32;
static oupfuncPtr oup32;

static void outbio32(unsigned char value, int port)
{
    (oup32)(port,value);
}

static unsigned char inbio32(int port)
{
    return (inp32)(port);
}

static __inline__ void outb(unsigned char _data, unsigned short _port)
{
    __asm__ __volatile__ ("outb %1, %0" : : "dN" (_port), "a" (_data));
}

static __inline__ unsigned char inb(unsigned short _port)
{
    unsigned char rv;
    __asm__ __volatile__ ("inb %1, %0" : "=a" (rv) : "dN" (_port));
    return rv;
}

static int getbio32(void);
static int sendbio32(unsigned char b);
#elif defined __FreeBSD__
static int iofd = -1;

static __inline unsigned char inb(unsigned short port) {
    unsigned char value;
    __asm __volatile("inb %%dx,%0" : "=a" (value) : "d" (port));
    return value;
}

static __inline void outb(unsigned char value, unsigned short port) {
    __asm __volatile("outb %0,%1" : : "a" (value), "id" (port));
}
#elif defined __DJGPP__
inline unsigned char inb(unsigned short _port)
{
    unsigned char rv;
    __asm__ __volatile__ ("inb %1, %0" : "=a" (rv) : "dN" (_port));
    return rv;
}

inline void outb(unsigned short _port, unsigned char _data)
{
    __asm__ __volatile__ ("outb %1, %0" : : "dN" (_port), "a" (_data));
}
#else
#define PARPORT

#include <sys/ioctl.h>
#include <fcntl.h>
#endif // WIN32

#ifdef PARPORT
static int parportfd = -1;
static int parportin;

#include <linux/ppdev.h>
#include <linux/parport.h>

static int getbparport(void);
static int sendbparport(unsigned char b);
#endif

static int vtime=1;

static int getb(void);
static int sendb(unsigned char b);

static const char *drivername(void) {
    return "IEC";
}

static int initialize(const char *dev, int port, Log *lf, int lastfail) {
    i_dev = dev; i_port = port;
    inited = 0;

    if (port & X1541_HOG) vtime=0;
    IOPORT = port & 0xffff;
    STATPORT = IOPORT + 1;
    CTRLPORT = IOPORT + 2;
    switch (port & X1541_MODE) {
    case X1541_X1541:
        DATAHI=0;DATALO=8;DATA=8;
        CLKHI=0;CLKLO=2;
        ATNHI=0;
        DATA_=8;CLK_=2;ATN_=1;
        inv=0;PORTIN=CTRLPORT;
#ifdef PARPORT
        parportin=PPRCONTROL;
#endif
        break;
    case X1541_XE1541:
        DATAHI=0;DATALO=8;DATA=8;
        CLKHI=0;CLKLO=2;
        ATNHI=0;
        DATA_=0x80;CLK_=0x20;ATN_=0x10;
        inv=-1;PORTIN=STATPORT;
#ifdef PARPORT
        parportin=PPRSTATUS;
#endif
        break;
    case X1541_XM1541:
        DATAHI=4;DATALO=0;DATA=4;
        CLKHI=0;CLKLO=2;
        ATNHI=0;
        DATA_=0x40;CLK_=0x20;ATN_=0x10;
        inv=-1;PORTIN=STATPORT;
#ifdef PARPORT
        parportin=PPRSTATUS;
#endif
        break;
    case X1541_XA1541:
        DATAHI=0;DATALO=4;DATA=4;
        CLKHI=2;CLKLO=0;
        ATNHI=1;
        DATA_=0x40;CLK_=0x20;ATN_=0x10;
        inv=-1;PORTIN=STATPORT;
#ifdef PARPORT
        parportin=PPRSTATUS;
#endif
        break;
    }
    x1541_io_ops.getb = getb;
    x1541_io_ops.sendb = sendb;
#ifdef WIN32
    do {
        /* Load the library */
        if (hLib == NULL) {
            hLib = LoadLibrary("inpout32.dll");
            if (hLib == NULL) {
                log_print(lf, "LoadLibrary of inpout32.dll failed, using direct access\n");
                break;
            } 
        }
        if (inp32 == NULL) {
            /* get the address of the function */
            inp32 = (inpfuncPtr) GetProcAddress(hLib, "Inp32");
            if (inp32 == NULL) {
                log_print(lf, "GetProcAddress for Inp32 failed, using direct access\n");
                break;
            }
        }
        if (oup32 == NULL) {
            oup32 = (oupfuncPtr) GetProcAddress(hLib, "Out32");
            if (oup32 == NULL) {
                log_print(lf, "GetProcAddress for Oup32 failed, using direct access\n");
                break;
            }
        }
        x1541_io_ops.getb = getbio32;
        x1541_io_ops.sendb = sendbio32;
    } while (0);
#elif defined __FreeBSD__
    iofd = open("/dev/io", O_RDWR);
    if (iofd == -1) {
        if (lastfail != -1) log_print(lf, "Error opening %s io device: %s(%d)\n", dev, strerror(errno), errno);
        return -1;
    }
#elif defined __DJGPP__
#else
#ifdef PARPORT
    parportfd = -1;
    if (port & X1541_DEVICE) {
        parportfd = open(dev, O_RDWR);
        if (parportfd == -1) {
            if (lastfail != -1) log_print(lf, "Error opening %s parport device: %s(%d)\n", dev, strerror(errno), errno);
            return -1;
        }
        if (ioctl(parportfd, PPEXCL) || ioctl(parportfd, PPCLAIM)) {
            close(parportfd);
            if (lastfail != -2) log_print(lf, "Cannot access %s parport device: %s(%d)\n", dev, strerror(errno), errno);
            return -2;
        }
        {
            int mode = IEEE1284_MODE_ECP;
            if (ioctl(parportfd, PPSETMODE, &mode)) {
                close(parportfd);
                if (lastfail != -3) log_print(lf, "Cannot set byte mode on %s parport device: %s(%d)\n", dev, strerror(errno), errno);
                return -3;
            }

            mode=0;
            if (ioctl(parportfd, PPDATADIR, &mode)) {
                close(parportfd);
                if (lastfail != -4) log_print(lf, "Cannot set direction on %s parport device: %s(%d)\n", dev, strerror(errno), errno);
                return -4;
            }
        }
        x1541_io_ops.getb=getbparport;
        x1541_io_ops.sendb=sendbparport;
    } else
#endif
    {
        if (ioperm(IOPORT,3,1)) {
            if (lastfail != -5) log_print(lf, "Cannot allocate I/O ports %x to %x: %s(%d)\n", IOPORT, IOPORT+2, strerror(errno), errno);
            return -5;
        }
    }
#endif
    inited = 1;
    return 0;
}

#ifdef WIN32
static inline int watnio32(int i) {
    unsigned char b;
    while (!tout) {
        b=inbio32(PORTIN);
        if (((b ^ inv) & ATN_)!=i) return 0;
    }
    return -EIO;
}

static int getbio32() {
    unsigned char b=0,c;
    outbio32(DATALO+CLKHI+ATNHI, CTRLPORT);
    alarm(1);

    if (watnio32(0)) return -EIO;
    if (inbio32(PORTIN) & CLK_) b|=128;
    outbio32(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watnio32(ATN_)) return -EIO;
    c=inbio32(PORTIN);
    if (c & CLK_) b|=64;
    outbio32(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnio32(0)) return -EIO;
    if (inbio32(PORTIN) & CLK_) b|=32;
    outbio32(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watnio32(ATN_)) return -EIO;
    c=inbio32(PORTIN);
    if (c & CLK_) b|=16;
    outbio32(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnio32(0)) return -EIO;
    if (inbio32(PORTIN) & CLK_) b|=8;
    outbio32(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watnio32(ATN_)) return -EIO;
    c=inbio32(PORTIN);
    if (c & CLK_) b|=4;
    outbio32(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnio32(0)) return -EIO;
    if (inbio32(PORTIN) & CLK_) b|=2;
    outbio32(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watnio32(ATN_)) return -EIO;
    c=inbio32(PORTIN);
    if (c & CLK_) b|=1;
    if ((c ^ inv) & DATA_) tout=1;
    b^=inv;
    crc_add(b);
    return b;
}

static int sendbio32(unsigned char a) {
    int b, c;
    b=a ^ (a << 1);
    crc_add(a);
    c=a & 128?(CLKHI | ATNHI | DATALO):(CLKLO | ATNHI | DATALO);
    outbio32(c ^ DATA, CTRLPORT);outbio32(c, CTRLPORT);
    alarm(1);
    if (b & 128) c^=CLK;
    if (watnio32(0)) return -EIO;
    if (b & 128) outbio32(c, CTRLPORT);
    outbio32(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (b & 64) c^=CLK;
    if (watnio32(ATN_)) return -EIO;
    if (b & 64) outbio32(c ^ DATA, CTRLPORT);
    outbio32(c, CTRLPORT);
    if (b & 32) c^=CLK;
    if (watnio32(0)) return -EIO;
    if (b & 32) outbio32(c, CTRLPORT);
    outbio32(c ^ DATA, CTRLPORT);
    if (b & 16) c^=CLK;
//    if (in() & DATA_) tout=1;
    if (watnio32(ATN_)) return -EIO;
    if (b & 16) outbio32(c ^ DATA, CTRLPORT);
    outbio32(c, CTRLPORT);
    if (b & 8) c^=CLK;
    if (watnio32(0)) return -EIO;
    if (b & 8) outbio32(c, CTRLPORT);
    outbio32(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (b & 4) c^=CLK;
    if (watnio32(ATN_)) return -EIO;
    if (b & 4) outbio32(c ^ DATA, CTRLPORT);
    outbio32(c, CTRLPORT);
    if (b & 2) c^=CLK;
    if (watnio32(0)) return -EIO;
    if (b & 2) outbio32(c, CTRLPORT);
    outbio32(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (watnio32(ATN_)) return -EIO;
    return 0;
}
#elif defined __FreeBSD__
#elif defined PARPORT
static inline int watnparport(int i) {
    unsigned char b;
    while (!tout) {
        ioctl(parportfd, parportin, &b);
        if (((b ^ inv) & ATN_)!=i) return 0;
    }
    return -EIO;
}

static int getbparport() {
    unsigned char b=0, dl=DATALO+CLKHI+ATNHI, dh=DATAHI+CLKHI+ATNHI, c;
    ioctl(parportfd, PPWCONTROL, &dl);
    alarm(1);

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=128;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=64;
    ioctl(parportfd, PPWCONTROL, &dl);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=32;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=16;
    ioctl(parportfd, PPWCONTROL, &dl);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=8;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=4;
    ioctl(parportfd, PPWCONTROL, &dl);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=2;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=1;
    if ((c ^ inv) & DATA_) tout=1;
    b^=inv;
    crc_add(b);
    return b;
}

static int sendbparport(unsigned char a) {
    int b, c;
    b=a ^ (a << 1);
    crc_add(a);
    c=a & 128?(CLKHI | ATNHI | DATAHI):(CLKLO | ATNHI | DATAHI);
    ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    alarm(1);
    if (b & 128) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 128) ioctl(parportfd, PPWCONTROL, &c);
    c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
//    if (in() & DATA_) tout=1;
    if (b & 64) c^=CLK;
    if (watnparport(ATN_)) return -EIO;
    if (b & 64) ioctl(parportfd, PPWCONTROL, &c);
    c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 32) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 32) ioctl(parportfd, PPWCONTROL, &c);
    c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 16) c^=CLK;
//    if (in() & DATA_) tout=1;
    if (watnparport(ATN_)) return -EIO;
    if (b & 16) ioctl(parportfd, PPWCONTROL, &c);
    c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 8) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 8) ioctl(parportfd, PPWCONTROL, &c);
    c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
//    if (in() & DATA_) tout=1;
    if (b & 4) c^=CLK;
    if (watnparport(ATN_)) return -EIO;
    if (b & 4) ioctl(parportfd, PPWCONTROL, &c);
    c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 2) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 2) ioctl(parportfd, PPWCONTROL, &c);
    c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
//    if (in() & DATA_) tout=1;
    if (watnparport(ATN_)) return -EIO;
    return 0;
}
#endif

static inline int watn(int i) {
    unsigned char b;
    while (!tout) {
        b=inb(PORTIN);
        if (((b ^ inv) & ATN_)!=i) return 0;
    }
    return -EIO;
}

static int getb() {
    unsigned char b=0,c;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    alarm(1);

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=128;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=64;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=32;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=16;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=8;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=4;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=2;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=1;
    if ((c ^ inv) & DATA_) tout=1;
    b^=inv;
    crc_add(b);
    return b;
}

static int sendb(unsigned char a) {
    int b, c;
    b=a ^ (a << 1);
    crc_add(a);
    c=a & 128?(CLKHI | ATNHI | DATALO):(CLKLO | ATNHI | DATALO);
    outb(c ^ DATA, CTRLPORT);outb(c, CTRLPORT);
    alarm(1);
    if (b & 128) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 128) outb(c, CTRLPORT);
    outb(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (b & 64) c^=CLK;
    if (watn(ATN_)) return -EIO;
    if (b & 64) outb(c ^ DATA, CTRLPORT);
    outb(c, CTRLPORT);
    if (b & 32) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 32) outb(c, CTRLPORT);
    outb(c ^ DATA, CTRLPORT);
    if (b & 16) c^=CLK;
//    if (in() & DATA_) tout=1;
    if (watn(ATN_)) return -EIO;
    if (b & 16) outb(c ^ DATA, CTRLPORT);
    outb(c, CTRLPORT);
    if (b & 8) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 8) outb(c, CTRLPORT);
    outb(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (b & 4) c^=CLK;
    if (watn(ATN_)) return -EIO;
    if (b & 4) outb(c ^ DATA, CTRLPORT);
    outb(c, CTRLPORT);
    if (b & 2) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 2) outb(c, CTRLPORT);
    outb(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (watn(ATN_)) return -EIO;
    return 0;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    int i, j;
    for (i=0;i<bytes;i++) {
        j=x1541_io_ops.getb();
        if (j<0) return j;
        data[i]=j;
    }
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    int i, j;
    for (i=0;i<bytes;i++) {
        j=x1541_io_ops.sendb(data[i]);
        if (j) return j;
    }
    return 0;
}

static void eshutdown(void) {
#ifdef WIN32
#elif defined __FreeBSD__
    if (iofd >= 0) {
        close(iofd);
        iofd = -1;
    }
#elif defined PARPORT
    if (parportfd >= 0) {
        ioctl(parportfd, PPRELEASE);
        close(parportfd);
        parportfd = -1;
    }
#endif
    inited = 0;
    return;
}

static int reinitialize(Log *lf, int lastfail) {
    eshutdown();
    return initialize(i_dev, i_port, lf, lastfail);
}

static int flush(void) {
    alarm(0);
    return 0;
}

static void clean(void) {
    return;
}

static int waitb(void) {
    int dyntime=1;
    int i;
start:
    alarm(0);
    if (!inited) return -ENODEV;
#ifdef PARPORT
    if (parportfd >= 0) {
        unsigned char c;
        c=CLKLO+DATAHI+ATNHI;ioctl(parportfd, PPWCONTROL, &c);
        for (;;) {
            ioctl(parportfd, parportin, &c);
            if ((c ^ inv) & DATA_) break;
            if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
        }
        c=CLKHI+DATAHI+ATNHI;ioctl(parportfd, PPWCONTROL, &c);
        tout=0;alarm(1);
        for (;;) {
            ioctl(parportfd, parportin, &c);
            if (((c ^ inv) & DATA_)==0) break;
            if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
            if (tout) goto start;
        } //set clklo, wait atnlo
    } else
#endif
    {
        outb(CLKLO+DATAHI+ATNHI, CTRLPORT);
        while (((inb(PORTIN) ^ inv) & DATA_)==0) if (vtime) {
            usleep(dyntime);
            if (dyntime<16384) dyntime<<=1;
        }
        outb(CLKHI+DATAHI+ATNHI, CTRLPORT);
        tout=0;alarm(1);
        while (((inb(PORTIN) ^ inv) & DATA_)!=0) {
            if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
            if (tout) goto start;
        } //set clklo, wait atnlo
    }
    i=x1541_io_ops.getb();
    if (tout) goto start;
    alarm(0);
    return (unsigned char)i;
}

struct io_ops_s x1541_io_ops = {
    .drivername   = drivername,
    .initialize   = initialize,
    .reinitialize = reinitialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = eshutdown,
    .flush        = flush,
    .wait         = waitb,
    .clean        = clean,
};
